# Decoder #

This decoder is the counterpart to the encoder. It is a program to translate the output of a MAX-SAT solver for a Cascading Portfolio Scheduling (CPS) instance 
back to CSP in a human-readable format. By default, the output is written to a .txt file in the 'results' directory.

The output consists of following parts:
- size of the trainingsset
- number of all available instances
- enumeration of the instances in the trainingsset by name
- runtime of the MAX-SAT solver in nanoseconds and seconds
- enumeration and comma separated list of the portfolio algorithms (by name and number) in the scheduled order
- costs of this schedule for the trainingsset
- costs of this schedule for all instances
- number of algorithms needed to solve the trainingsset
- number of algorithms needed to solve all instances

Optionally:
- schedule according to the greedy algorithm
- costs of the greedy schedule for the trainingsset
- costs of the greedy schedule for all instances
- number of algorithms needed to solve the trainingsset based on the greedy schedule
- number of algorithms needed to solve all instances based on the greedy schedule

## Setup and running ##

### 1. Build ###

Build the project using Maven:

```
mvn clean install
```

_Prerequisite:_  Maven 3.6.x or higher and Java 10 have to be available on your computer.

### 2. Execution ###

The decoder can be run in 3 different ways depending on the number of arguments passed to Main.java:
- 2 arguments
- 6 arguments
- 7 arguments

#### With 2 Arguments ####

The first argument has to be the same .csv file with the time/cost information about the portfolio that was used for the encoding.

The seconds argument has to be the same .csv file with the solving information about the portfolio that was used for the encoding.

Example for arguments:

1. decoder/src/main/resources/qdimacs-all_time.csv 
2. decoder/src/main/resources/qdimacs-all_solved.csv

Calling the decoder with 2 arguments is used for automatically running through certain directories to read and filter 
the remaining files that are necessary for execution (as explained in the next sections for 6 and 7 arguments). 
The program is based on the file structure in the 'resources' directory, but the strings in lines 387 and 389 have to be 
adapted to tell the decoder which subdirectories (instance number, MAX-SAT solver) to search:
```
String runType = "200_instances-20_solvers";
String maxSatSolverName = "OpenWBO";
```

To guide the decoder to your own directories and to specify where the output should be written to, you have to adapt the strings in lines 378 - 401.

#### With 6 Arguments ####

Execution of the decoder with 6 arguments is used to decode one specific output of a MAX-SAT solver.

The first argument is the output of the MAX-SAT solver.

The second argument is the .csv file starting with "Wcnf-" as produced by the encoder during the encoding for the MAX-SAT solver.

The third argument is the .csv file starting with "solver2id-" as produced by the encoder.

The fourth argument is the .csv file starting with "instance2id-" as produced by the encoder.

The fifth argument has to be the same .csv file with the time/cost information about the portfolio that was used for the encoding.

The sixth argument has to be the same .csv file with the solving information about the portfolio that was used for the encoding.

Example for arguments:

1. decoder/src/main/resources/solver-output/200_instances-20_solvers/maxino/maxino-output-Wcnf-2020-02-19-18-52-13.txt 
2. decoder/src/main/resources/encoding_variable-2-solver_variable-mapping/200_instances-20_solvers/Wcnf-2020-02-19-18-52-13.csv 
3. decoder/src/main/resources/solver2id-mapping/200_instances-20_solvers/solver2id-2020-02-19-18-52-13.csv 
4. decoder/src/main/resources/instance2id-mapping/200_instances-20_solvers/instance2id-2020-02-19-18-52-13.csv 
5. decoder/src/main/resources/qdimacs-all_time.csv 
6. decoder/src/main/resources/qdimacs-all_solved.csv


#### With 7 Arguments ####

Like an execution with 6 arguments, execution of the decoder with 7 arguments is used to decode one specific output of a MAX-SAT solver. 
The additional input comprises the output of the encoder for the greedy algorithm in order to get a comparison of the optimal solution by the MAX-SAT solver and the greedy solution.

The first argument is the output of the MAX-SAT solver.

The second argument is the .csv file starting with "Wcnf-" as produced by the encoder during the encoding for the MAX-SAT solver.

The third argument is the .csv file starting with "solver2id-" as produced by the encoder.

The fourth argument is the .csv file starting with "instance2id-" as produced by the encoder.

The fifth argument has to be the same .csv file with the time/cost information about the portfolio that was used for the encoding.

The sixth argument has to be the same .csv file with the solving information about the portfolio that was used for the encoding.

The seventh argument is the .csv file with the output of the greedy scheduling (starting with "Greedy-") as produced by the encoder.

Example for arguments:

1. decoder/src/main/resources/solver-output/200_instances-20_solvers/maxino/maxino-output-Wcnf-2020-02-19-18-52-13.txt 
2. decoder/src/main/resources/encoding_variable-2-solver_variable-mapping/200_instances-20_solvers/Wcnf-2020-02-19-18-52-13.csv 
3. decoder/src/main/resources/solver2id-mapping/200_instances-20_solvers/solver2id-2020-02-19-18-52-13.csv 
4. decoder/src/main/resources/instance2id-mapping/200_instances-20_solvers/instance2id-2020-02-19-18-52-13.csv 
5. decoder/src/main/resources/qdimacs-all_time.csv 
6. decoder/src/main/resources/qdimacs-all_solved.csv
7. decoder/src/main/resources/greedy-output/200_instances-20_solvers/Greedy-2020-02-19-18-52-13.csv

## Copyright ##

The copyright for the program lies with the author, Maria Bresich, but it can be used for academic purposes.

