# Encoder #

This encoder is the counterpart to the decoder. It is a program to not only translate a Cascading Portfolio Scheduling (CPS) instance 
to a MAX-SAT instance but also to solve the CPS instance with two different Branch-and-Bound (B&B) based algorithms and one greedy algorithm. 
By default, the output files include the current timestamp in their names and are written to the 'results' directory.

Following output files are generated:
- MAX-SAT encoded CPS instance (.wcnf file starting with "Wcnf-")
- MAX-SAT encoded CPS instance for solvers that can deal with floating point weights (.wcnf file starting with "Float-Wcnf-")
- mapping file of the variables used for the encoding to the numbered variables used in the .wcnf file as input for MAX-SAT solvers (.csv file starting with "Wcnf-")
- .csv file with the runtime of the encoding procedure (starting with "Encoding-")
- mapping file of instance names to IDs (.csv file starting with "instance2id-")
- mapping file of algorithm names to IDs (.csv file starting with "solver2id-")
- output of a plain B&B algorithm (.csv file starting with "B&B-")
- output of an optimized B&B algorithm (.csv file starting with "B&B-Opt-")
- output of the greedy algorithm (.csv file starting with "Greedy-")

The mapping files are needed for the translation of the output of MAX-SAT solvers back to CPS using the decoder program.

## Setup and running ##

### 1. Build ###

Build the project using Maven:

```
mvn clean install
```

_Prerequisite:_  Maven 3.6.x or higher and Java 10 have to be available on your computer.

### 2. Execution ###

The encoder can be run in 3 different ways depending on the number of arguments passed to Test.java:
- 0 arguments
- 2 arguments
- 5 arguments

#### With 0 Arguments ####

In case that the encoder is executed with no arguments, it runs with a randomly generated CPS instance. By default, a 
portfolio of 4 algorithms and a trainingsset of 10 instances is generated. These settings can be changed in lines 52 and 53:

```
int noOfAlgorithms = 4;
int noOfInstances = 10;
```

#### With 2 Arguments ####

Execution with 2 arguments is the main use case for the encoder where the input constitutes a CPS instance.

The first argument has to be a .csv file with the time/cost information about the portfolio.

The second argument has to be a .csv file with the solving information about the portfolio.

Example for arguments:

1. encoder/src/main/resources/qdimacs-all_time.csv 
2. encoder/src/main/resources/qdimacs-all_solved.csv

By default, the program will randomly pick 200 test instances for the trainingsset and 20 algorithms from the portfolio. 
These values can be changed by editing lines 124 and 125:

```
int noOfPickedInstances = 200;
int noOfPickedSolvers = 20;
```


#### With 5 Arguments ####

Execution of the encoder with 5 arguments can be used to rerun the B&B and/or greedy algorithms for an CPS instance. 
The input data is needed to retrace which test instances were originally used for the trainingsset and which algorithms were picked from the portfolio.

The first argument is the .csv file starting with "Wcnf-" as originally produced by the encoder during the encoding.

The second argument is the .csv file starting with "solver2id-" as produced by the encoder.

The third argument is the .csv file starting with "instance2id-" as produced by the encoder.

The fourth argument has to be the same .csv file with the time/cost information about the portfolio that was originally used for the encoding.

The fifth argument has to be the same .csv file with the solving information about the portfolio that was originally used for the encoding.

Example for arguments:

1. encoder/results/Wcnf-2020-09-30-22-18-08.csv 
2. encoder/results/solver2id-2020-09-30-22-18-08.csv 
3. encoder/results/instance2id-2020-09-30-22-18-08.csv 
4. encoder/src/main/resources/qdimacs-all_time.csv 
5. encoder/src/main/resources/qdimacs-all_solved.csv

By default, both B&B algorithms and the greedy algorithm are run again. This behavior can be changed by in-/excluding lines 245, 252, and 264 in the MyParser.java file where the respective algorithms are called.

## Copyright ##

The copyright for the program lies with the author, Maria Bresich, but it can be used for academic purposes.
