import java.io.*;
import java.nio.charset.StandardCharsets;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;

public class MyBranchAndBound implements Runnable {

    private String resultDirectoryName;
    private String fileName;
    private StringBuilder stringBuilder;

    private long[][] solved;
    private long[][] costs;
    private float[][] floatCosts;

    private long upperBound;
    private long lowerBound;

    private int numberOfSolvers;
    private int numberOfInstances;

    List<List<Integer>> instancesSolvedBySolvers;
    List<List<Integer>> solverSolvesInstances;

    private long[] overallCostsPerSolver;
    private List<Integer> totalOrder = new LinkedList<>();
    private List<Integer> currentBestSolution = new LinkedList<>();


    public MyBranchAndBound() {
        this.instancesSolvedBySolvers = new ArrayList<List<Integer>>(this.numberOfInstances);
        this.solverSolvesInstances = new ArrayList<List<Integer>>(this.numberOfSolvers);

        for (int i = 0; i < this.numberOfInstances; i++) {
            this.instancesSolvedBySolvers.add(new LinkedList<Integer>());
        }

        for (int i = 0; i < this.numberOfSolvers; i++) {
            this.solverSolvesInstances.add(new LinkedList<Integer>());
        }

        this.overallCostsPerSolver = new long[this.numberOfSolvers];

    }

    public MyBranchAndBound(int numberOfSolvers, int numberOfInstances) {
        this.numberOfSolvers = numberOfSolvers;
        this.numberOfInstances = numberOfInstances;

        this.instancesSolvedBySolvers = new ArrayList<List<Integer>>(numberOfInstances);
        this.solverSolvesInstances = new ArrayList<List<Integer>>(numberOfSolvers);

        for (int i = 0; i < numberOfInstances; i++) {
            this.instancesSolvedBySolvers.add(new LinkedList<Integer>());
        }

        for (int i = 0; i < numberOfSolvers; i++) {
            this.solverSolvesInstances.add(new LinkedList<Integer>());
        }

        this.overallCostsPerSolver = new long[numberOfSolvers];

    }

    public MyBranchAndBound(String resultDirectoryName, String uniqueTime, int numberOfSolvers, int numberOfInstances) {
        this.resultDirectoryName = resultDirectoryName;
        this.fileName = "B&B-" + uniqueTime + ".csv";
        this.stringBuilder = new StringBuilder();

        this.numberOfSolvers = numberOfSolvers;
        this.numberOfInstances = numberOfInstances;

        this.instancesSolvedBySolvers = new ArrayList<List<Integer>>(numberOfInstances);
        this.solverSolvesInstances = new ArrayList<List<Integer>>(numberOfSolvers);

        for (int i = 0; i < numberOfInstances; i++) {
            this.instancesSolvedBySolvers.add(new LinkedList<Integer>());
        }

        for (int i = 0; i < numberOfSolvers; i++) {
            this.solverSolvesInstances.add(new LinkedList<Integer>());
        }

        this.overallCostsPerSolver = new long[numberOfSolvers];

    }

    private void initializeArrays() {
        this.solved = new long[this.numberOfSolvers][this.numberOfInstances];
        this.costs = new long[this.numberOfSolvers][this.numberOfInstances];

        this.solved[0][0] = 1;
        this.solved[0][1] = 1;
        this.solved[0][2] = 1;
        this.solved[0][3] = 0;
        this.solved[0][4] = 1;

        this.solved[1][0] = 0;
        this.solved[1][1] = 0;
        this.solved[1][2] = 1;
        this.solved[1][3] = 0;
        this.solved[1][4] = 1;

        this.solved[2][0] = 0;
        this.solved[2][1] = 1;
        this.solved[2][2] = 0;
        this.solved[2][3] = 1;
        this.solved[2][4] = 0;

        this.solved[3][0] = 1;
        this.solved[3][1] = 1;
        this.solved[3][2] = 1;
        this.solved[3][3] = 0;
        this.solved[3][4] = 1;

        this.costs[0][0] = 1;
        this.costs[0][1] = 5;
        this.costs[0][2] = 2;
        this.costs[0][3] = 7;
        this.costs[0][4] = 3;

        this.costs[1][0] = 7;
        this.costs[1][1] = 7;
        this.costs[1][2] = 3;
        this.costs[1][3] = 7;
        this.costs[1][4] = 5;

        this.costs[2][0] = 7;
        this.costs[2][1] = 1;
        this.costs[2][2] = 7;
        this.costs[2][3] = 6;
        this.costs[2][4] = 7;

        this.costs[3][0] = 2;
        this.costs[3][1] = 5;
        this.costs[3][2] = 3;
        this.costs[3][3] = 7;
        this.costs[3][4] = 4;
    }

    private void fillLists() {
        for (int i = 0; i < this.numberOfSolvers; i++) {
            for (int j = 0; j < this.numberOfInstances; j++) {
                if (this.solved[i][j] == 1) {
                    this.solverSolvesInstances.get(i).add(j);
                    this.instancesSolvedBySolvers.get(j).add(i);
                }
            }
        }
    }

    private void computeOverallCostsForSolvers() {
        for (int i = 0; i < this.numberOfSolvers; i++) {
            for (int j = 0; j < this.numberOfInstances; j++) {
                this.overallCostsPerSolver[i] += this.costs[i][j];
            }
        }
    }

    private void computeInitialUpperBound() {
        for (int i = 0; i < this.numberOfSolvers; i++) {
            for (int j = 0; j < this.numberOfInstances; j++) {
                this.upperBound += this.costs[i][j];
            }
        }
        //System.out.println("Initial upper bound: " + upperBound);
    }


    // nur Kopie von solution-Liste übergeben, Liste wird in der Methode verändert
    private long computeUpperBound(List<Integer> currentSolution) {
        long currentCosts = 0;
        int oldSolutionSize = currentSolution.size();
        List<Integer> notSolvedInstances = new LinkedList<>();
        for (int i = 0; i < this.numberOfInstances; i++) {
            notSolvedInstances.add(i);
        }

        //real costs for current solution
        if (!currentSolution.isEmpty()) {
            for (int i = 0; i < currentSolution.size(); i++) {
                int currentSolver = currentSolution.get(i);
                if (i == 0) {       //sollte eigentlich vom else-Zweig abgedeckt sein
                    currentCosts += this.overallCostsPerSolver[currentSolver];
                    for (int j = 0; j < this.solverSolvesInstances.get(currentSolver).size(); j++) {
                        int instance = this.solverSolvesInstances.get(currentSolver).get(j);
                        if (notSolvedInstances.contains(instance)) {
                            ((LinkedList<Integer>) notSolvedInstances).removeFirstOccurrence(instance);
                        }
                    }
                } else {
                    for (int j = 0; j < notSolvedInstances.size(); j++) {
                        currentCosts += this.costs[currentSolver][notSolvedInstances.get(j)];
                    }
                    for (int j = 0; j < this.solverSolvesInstances.get(currentSolver).size(); j++) {
                        int instance = this.solverSolvesInstances.get(currentSolver).get(j);
                        if (notSolvedInstances.contains(instance)) {
                            ((LinkedList<Integer>) notSolvedInstances).removeFirstOccurrence(instance);

                        }
                    }
                }
            }
        }

        //costs for all remaining instances
        for (int i = 0; i < notSolvedInstances.size(); i++) {
            int instance = notSolvedInstances.get(i);
            //System.out.println("not solved: " + instance);
            if (!this.instancesSolvedBySolvers.get(instance).isEmpty()) {
                int solver = this.instancesSolvedBySolvers.get(instance).get(0);     //get first solving solver, not sorted
                if (!currentSolution.contains(solver)) {
                    currentSolution.add(solver);
                }
            }
        }

        for (int i = oldSolutionSize; i < currentSolution.size(); i++) {
            int solver = currentSolution.get(i);
            //add according costs
            for (int j = 0; j < notSolvedInstances.size(); j++) {
                currentCosts += this.costs[solver][notSolvedInstances.get(j)];
            }
            //remove instances solved by this new solver in the solution from the list -> to get a lower upper bound
            for (int j = 0; j < this.solverSolvesInstances.get(solver).size(); j++) {
                int instanceToRemove = this.solverSolvesInstances.get(solver).get(j);
                if (notSolvedInstances.contains(instanceToRemove)) {
                    ((LinkedList<Integer>) notSolvedInstances).removeFirstOccurrence(instanceToRemove);

                }
            }

        }

        setCurrentBestSolution(currentSolution);
        return currentCosts;
    }

    private void setCurrentBestSolution(List<Integer> currentSolution) {
        this.currentBestSolution = currentSolution;
    }

    private void computeLowerBound(List<Integer> currentSolution) {
        List<Long> costList = new LinkedList<Long>();
        long tempLowerBound = 0;
        this.lowerBound = 0;    // reset global lower bound

        long currentCosts = 0;
        int oldSolutionSize = currentSolution.size();
        List<Integer> notSolvedInstances = new LinkedList<>();
        for (int i = 0; i < this.numberOfInstances; i++) {
            notSolvedInstances.add(i);
        }

        //real costs for current solution
        if (!currentSolution.isEmpty()) {
            for (int i = 0; i < currentSolution.size(); i++) {
                int currentSolver = currentSolution.get(i);
                if (i == 0) {       //sollte eigentlich vom else-Zweig abgedeckt sein
                    currentCosts += this.overallCostsPerSolver[currentSolver];
                    for (int j = 0; j < this.solverSolvesInstances.get(currentSolver).size(); j++) {
                        int instance = this.solverSolvesInstances.get(currentSolver).get(j);
                        if (notSolvedInstances.contains(instance)) {
                            ((LinkedList<Integer>) notSolvedInstances).removeFirstOccurrence(instance);
                        }
                    }
                } else {
                    for (int j = 0; j < notSolvedInstances.size(); j++) {
                        currentCosts += this.costs[currentSolver][notSolvedInstances.get(j)];
                    }
                    for (int j = 0; j < this.solverSolvesInstances.get(currentSolver).size(); j++) {
                        int instance = this.solverSolvesInstances.get(currentSolver).get(j);
                        if (notSolvedInstances.contains(instance)) {
                            ((LinkedList<Integer>) notSolvedInstances).removeFirstOccurrence(instance);

                        }
                    }
                }
            }
        }

        this.lowerBound += currentCosts;

        //costs for all remaining instances
        for (int i = 0; i < notSolvedInstances.size(); i++) {
            int instance = notSolvedInstances.get(i);
            costList.clear();

            for (int j = 0; j < this.numberOfSolvers; j++) {
                if (this.solved[j][instance] == 1) {
                    costList.add(this.costs[j][instance]);
                }
            }
            if (!costList.isEmpty()) {
                tempLowerBound = costList.get(0);
                for (int j = 0; j < costList.size(); j++) {
                    if (costList.get(j) < tempLowerBound) {
                        tempLowerBound = costList.get(j);
                    }
                }
                this.lowerBound += tempLowerBound;
            }
        }

        //System.out.println("Lower bound: " + lowerBound);
    }

    private void branchAndBound(int solver, List<Integer> solution) {
        if (solver < this.numberOfSolvers) {
            this.lowerBound = 0;
            // create copy of the current solution because the list will be changed during the computation of the lower bound
            List<Integer> solutionCopyLower = new LinkedList<>();
            solutionCopyLower.addAll(solution);
            computeLowerBound(solutionCopyLower);   // changes the value of the global lower bound which is stored in lowerBound

            //otherwise branching not necessary
            if (this.lowerBound < this.upperBound) {

                List<Integer> solutionCopy = new LinkedList<>();
                solutionCopy.addAll(solution);

                long localUpper = computeUpperBound(solutionCopy);

                if (localUpper < this.upperBound) {
                    this.upperBound = localUpper;
                    //System.out.println("New upper bound: " + upperBound);
                    //System.out.println("Current solution: " + solution.toString());
                    //System.out.println("Current best solution: " + this.currentBestSolution);

                    // totalOrder stores the order of the solvers of the currently global best solution
                    this.totalOrder.clear();
                    this.totalOrder.addAll(currentBestSolution);
                }

                if (this.lowerBound < this.upperBound) {
                    // Partitioning

                    for (int i = 0; i < (this.numberOfSolvers - solution.size()); i++) {

                        List<Integer> remainingPossibleSolvers = new LinkedList<>();
                        for (int j = 0; j < this.numberOfSolvers; j++) {
                            if (!solution.contains(j)) {
                                remainingPossibleSolvers.add(j);                            }
                        }

                        int nextSolver = ((LinkedList<Integer>) remainingPossibleSolvers).get(i);
                        solution.add(nextSolver);
                        //System.out.println("Current solution before next b&b-call: " + solution.toString());
                        branchAndBound(nextSolver, solution);
                        ((LinkedList<Integer>) solution).removeFirstOccurrence(nextSolver);

                    }

                }

            }
        }

    }


    @Override
    public void run() {
        long startTime = System.nanoTime();

        //initializeArrays();
        fillLists();
        computeOverallCostsForSolvers();

        computeInitialUpperBound();

        List<Integer> solution = new LinkedList<>();
        if (this.numberOfInstances >= 0 && this.numberOfSolvers >= 0) {
            for (int i = 0; i < this.numberOfSolvers; i++) {
                //System.out.println("First solver in order: " + i);
                solution.add(i);
                branchAndBound(i, solution);
                ((LinkedList<Integer>) solution).removeFirstOccurrence(i);

            }

            long endTime = System.nanoTime();
            double elapsedTime = (double) endTime - startTime;
            double elapsedTimeInSeconds = elapsedTime * 0.000000001;

            this.stringBuilder.append("runtime-in-nanoseconds").append(",").append(elapsedTime).append(System.lineSeparator());
            this.stringBuilder.append("runtime-in-seconds").append(",").append(elapsedTimeInSeconds).append(System.lineSeparator());
            this.stringBuilder.append("costs").append(",").append(this.upperBound).append(System.lineSeparator());
            this.stringBuilder.append("order").append(",");
            // add 1 to each solver to get the correct id (which start at 1, not 0)
            for (int i = 0; i < this.totalOrder.size(); i++) {
                if (i == (this.totalOrder.size() - 1)) {
                    this.stringBuilder.append(this.totalOrder.get(i)+1).append(System.lineSeparator());
                } else {
                    this.stringBuilder.append(this.totalOrder.get(i)+1).append(",");
                }
            }
            this.stringBuilder.append("number-needed-solvers").append(",").append(this.totalOrder.size()).append(System.lineSeparator());

            writeFile();

            System.out.println("Minimum costs: " + this.upperBound);
            System.out.println("Order of solvers: ");
            for (int i = 0; i < this.totalOrder.size(); i++) {
                System.out.print((this.totalOrder.get(i) + 1) + " ");
            }
            System.out.println("\n");
        }

    }

    private void writeFile() {
        try {
            File dir = new File(resultDirectoryName);
            File file = new File(dir, this.fileName);

            BufferedWriter writer = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(file), StandardCharsets.UTF_8));
            writer.append(this.stringBuilder);
            writer.flush();

        } catch (FileNotFoundException e) {
            e.printStackTrace();
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    public void setSolved(long[][] solved) {
        this.solved = solved;
    }

    public long[][] getSolved() {
        return this.solved;
    }

    public void setCosts(long[][] costs) {
        this.costs = costs;
    }

    public long[][] getCosts() {
        return costs;
    }

    public void setNumberOfSolvers(int numberOfSolvers) {
        this.numberOfSolvers = numberOfSolvers;
    }

    public void setNumberOfInstances(int numberOfInstances) {
        this.numberOfInstances = numberOfInstances;
    }

    public long getUpperBound() {
        return this.upperBound;
    }
}
