import java.io.*;
import java.nio.charset.StandardCharsets;
import java.util.*;

public class MyGreedyPortfolioScheduling implements Runnable{

    private String resultDirectoryName;
    private String fileName;
    private StringBuilder stringBuilder;

    private boolean randomInput = false;

    private long[][] solved;
    private long[][] costs;

    private int noOfPickedSolvers;
    private int noOfPickedInstances;

    private int noOfAllSolvers;
    private int noOfAllInstances;

    private String[] solverNames;
    private String[] instanceNames;

    private long[][] allSolved;
    private long[][] allCosts;

    private SortedMap<String, Integer> solver2id;
    private SortedMap<String, Integer> instance2id;

    private List<List<String>> costFileData;
    private List<List<String>> solveFileData;

    private int[] numberOfSolvedInstances;
    //private int[] costsForUnsolvedInstancesPerSolver;
    private long[] overallCostsPerSolver;
    private double[] overallCostsDividedByNumberOfSolvedInstances;

    //private static int NUMBER_OF_SOLVERS = 4;
    //private static int NUMBER_OF_INSTANCES = 5;

    private List<Integer> solversToSort = new LinkedList<>();
    private List<Integer> solversSortedByNumberOfSolvedInstances = new LinkedList<>();
    private List<Integer> solversSortedByAverageCostPerSolvedInstance = new LinkedList<>();
    private List<Integer> solversSortedByFirstSolvingSover = new LinkedList<>();
    private List<Integer> solversSortedByTotalCostsDividedByNumberOfSolvedInstances = new LinkedList<>();
    private List<Integer> totalOrder = new LinkedList<>();
    private List<String> totalOrderByNames = new LinkedList<>();

    MyGreedyPortfolioScheduling(String resultDirectoryName, String uniqueTime) {
        this.resultDirectoryName = resultDirectoryName;
        this.fileName = "Greedy-" + uniqueTime + ".csv";
        this.stringBuilder = new StringBuilder();
    }

    public void setDataForAllInstances(String[] solverNames, String[] instanceNames, List<List<String>> costFileData, List<List<String>> solveFileData) {
        this.solverNames = solverNames;
        this.instanceNames = instanceNames;

        this.costFileData = costFileData;
        this.solveFileData = solveFileData;

        this.noOfAllSolvers = costFileData.get(0).size() - 1;    // -1 because first entry is just "instance" and not a solver name
        this.noOfAllInstances = costFileData.size() - 1;    // -1 because the first row is not an instance but the name of the solvers

        this.allSolved = new long[this.noOfAllSolvers][this.noOfAllInstances];
        this.allCosts = new long[this.noOfAllSolvers][this.noOfAllInstances];

        // start with 1 because first entry is neither an algorithm nor an instance
        for (int i = 1; i <= this.noOfAllSolvers; i++) {
            for (int j = 1; j <= this.noOfAllInstances; j++) {
                this.allSolved[i - 1][j - 1] = Long.parseLong(solveFileData.get(j).get(i));
                this.allCosts[i - 1][j - 1] = (long) (Double.parseDouble(costFileData.get(j).get(i)) * 1000);    // convert cut time to milliseconds!! -> weight must be an int/long greater than 0!
            }
        }
    }

    private long[] computeCostsOfUnsolvedInstancesPerSolver(List<Integer> unsolvedInstances) {
        long[] costsForUnsolvedInstancesPerSolver = new long[this.noOfPickedSolvers];
        Arrays.fill(costsForUnsolvedInstancesPerSolver, Long.MAX_VALUE);

        for (int i = 0; i < this.solversToSort.size(); i++) {
            int solverId = solversToSort.get(i);
            costsForUnsolvedInstancesPerSolver[solverId] = 0;
            for (int j = 0; j < unsolvedInstances.size(); j++) {
                int instanceId = unsolvedInstances.get(j);

                costsForUnsolvedInstancesPerSolver[solverId] += this.costs[solverId][instanceId];
            }
        }

        return costsForUnsolvedInstancesPerSolver;
    }

    private List<Integer> sortSolversByNumberOfSolvedInstances(List<Integer> unsolvedInstances) {

        List<Integer> sortedSolvers = new LinkedList<>();

        if (this.numberOfSolvedInstances == null) {
            this.numberOfSolvedInstances = new int[this.noOfPickedSolvers];
        }

        // reset number
        Arrays.fill(this.numberOfSolvedInstances, 0);

        for (int i = 0; i < this.solversToSort.size(); i++) {
            int solverId = this.solversToSort.get(i);
            for (int j = 0; j < unsolvedInstances.size(); j++) {
                int unsolvedInstanceId = unsolvedInstances.get(j);

                if (this.solved[solverId][unsolvedInstanceId] == 1) {
                    this.numberOfSolvedInstances[solverId]++;
                }
            }
        }


        long[] costsOfUnsolvedInstancesPerSolver = computeCostsOfUnsolvedInstancesPerSolver(unsolvedInstances);

        double[] quotientSolvedDividedByCosts = new double[this.noOfPickedSolvers];
        for (int i = 0; i < this.noOfPickedSolvers; i++) {
            quotientSolvedDividedByCosts[i] = this.numberOfSolvedInstances[i] / (double) costsOfUnsolvedInstancesPerSolver[i];
        }


        // sorting
        // if same number of solved instances, then sort by increasing costs

        for (int i = 0; i < this.solversToSort.size(); i++) {
            int currentSolver = this.solversToSort.get(i);
            if (sortedSolvers.isEmpty()) {
                sortedSolvers.add(currentSolver);
            } else {
                int index = 0;
                boolean put = false;
                while (!put) {
                    if (index == sortedSolvers.size()) {    // reached end of the list
                        sortedSolvers.add(currentSolver);
                        put = true;
                    } else if (quotientSolvedDividedByCosts[currentSolver] > quotientSolvedDividedByCosts[sortedSolvers.get(index)]) {
                        sortedSolvers.add(index, currentSolver);
                        put = true;
                    } else if (quotientSolvedDividedByCosts[currentSolver] == quotientSolvedDividedByCosts[sortedSolvers.get(index)]) {

                        // 2nd criterium: costs are smaller -> alternative: number of solved is bigger = better?
                        if (costsOfUnsolvedInstancesPerSolver[currentSolver] <= costsOfUnsolvedInstancesPerSolver[sortedSolvers.get(index)]) {    // check costs
                            sortedSolvers.add(index, currentSolver);
                            put = true;
                        } else {
                            sortedSolvers.add(index+1, currentSolver);
                            put = true;
                        }
                    } else {
                        index++;
                    }
                }
            }
        }

        return sortedSolvers;
    }

    private void computeTotalOrderCosts() {
        long costs = 0;
        boolean[][] alreadySolved = new boolean[this.noOfPickedSolvers][this.noOfPickedInstances];

        for (Integer i : this.totalOrder) {  // i corresponds to the number of the algorithm which start at 1 => use i-1 for arrays as they start with index 0
            for (int j = 0; j < this.noOfPickedInstances; j++) {
                if (!alreadySolved[i - 1][j]) {
                    costs += (long)this.costs[i - 1][j];
                    if (this.solved[i - 1][j] == 1) {
                        for (int k = 0; k < this.noOfPickedSolvers; k++) {
                            alreadySolved[k][j] = true;
                        }
                    }
                }
            }
        }

        System.out.println("Costs of this order: " + costs);
        this.stringBuilder.append("costs").append(",").append(costs).append(System.lineSeparator());
    }

    private void computeTotalOrderCostsForRandomInput() {
        long costs = 0;
        boolean[][] alreadySolved = new boolean[this.noOfPickedSolvers][this.noOfPickedInstances];

        for (Integer i : this.totalOrder) {
            for (int j = 0; j < this.noOfPickedInstances; j++) {
                if (!alreadySolved[i - 1][j]) {     // -1 because total order contains solver numbers and not indices
                    costs += (long)this.costs[i - 1][j];
                    if (this.solved[i - 1][j] == 1) {
                        for (int k = 0; k < this.noOfPickedSolvers; k++) {
                            alreadySolved[k][j] = true;
                        }
                    }
                }
            }
        }
        System.out.println("Costs of this order: " + costs);
        this.stringBuilder.append("costs").append(",").append(costs).append(System.lineSeparator());
    }

    private void checkAllOrdered() {
        List<Integer> unsolvedInstances = new LinkedList<>();
        for (int i = 0; i < this.noOfPickedInstances; i++) {
            unsolvedInstances.add(i);
        }

        List<Integer> solversSorted = sortSolversByNumberOfSolvedInstances(unsolvedInstances);

        // add the remaining solvers according to the sorting considering all instances
        if (this.randomInput) {
            if (totalOrder.size() != this.noOfPickedSolvers) {
                for (int i = 0; i < solversSorted.size(); i++) {
                    if (this.solversToSort.contains(solversSorted.get(i))) {
                        this.totalOrder.add(solversSorted.get(i) + 1);
                    }
                }
            }
        } else {
            if (totalOrder.size() != this.noOfPickedSolvers) {
                for (int i = 0; i < solversSorted.size(); i++) {
                    if (this.solversToSort.contains(solversSorted.get(i))) {
                        this.totalOrder.add(solversSorted.get(i) + 1);
                    }
                }
            }
        }

    }

    private void computeCostsOfThisOrderForAllInstances() {
        long orderCosts = 0;
        long checkCorrectCost = 0;
        int enterOrig = 0;
        int enterNew = 0;
        boolean[][] alreadySolved = new boolean[this.noOfPickedSolvers][this.noOfAllInstances];
        boolean[][] alreadySolvedCheck = new boolean[this.noOfPickedSolvers][this.noOfAllInstances];


        // you have to look up the original id! because allCosts and allSolved entries correspond to the original ids
        for (Integer i : this.totalOrder) {  // i corresponds to the number of the algorithm which start at 1 => use i-1 for arrays as they start with index 0
            //test correctness - start
            String chosenSolverName = null;
            for (Map.Entry<String, Integer> entry : this.solver2id.entrySet()) {
                if (i == entry.getValue()) {
                    chosenSolverName = entry.getKey();
                }
            }

            int solverOriginalId = -1;
            for (int k = 1; k < this.costFileData.get(0).size(); k++) {     // first entry is not a solver
                if (this.costFileData.get(0).get(k).equals(chosenSolverName)) {
                    //System.out.println("cost file data entry = " + this.costFileData.get(0).get(k));
                    solverOriginalId = k;
                    //System.out.println("chosen solver name: " + chosenSolverName);
                    //System.out.println("original id = " + solverOriginalId + " vs current id = " + i);
                }
            }

            // test correctness - end

            for (int j = 0; j < this.noOfAllInstances; j++) {

                //test correctness - start

                if (!alreadySolvedCheck[i - 1][j]) {
                    enterOrig++;
                    checkCorrectCost += (long)this.allCosts[solverOriginalId - 1][j];
                    //System.out.println("original id - added costs = " + this.allCosts[solverOriginalId - 1][j]);
                    if (this.allSolved[solverOriginalId - 1][j] == 1) {
                        for (int k = 0; k < this.noOfPickedSolvers; k++) {
                            alreadySolvedCheck[k][j] = true;
                        }
                    }
                }

                // test correctness - end

                // correctness test above showed that this is not correct when not all available solvers are picked
                if (!alreadySolved[i - 1][j]) {
                    enterNew++;
                    orderCosts += (long)this.allCosts[i - 1][j];
                    //System.out.println("new id - added costs = " + this.allCosts[i - 1][j]);
                    if (this.allSolved[i - 1][j] == 1) {
                        for (int k = 0; k < this.noOfPickedSolvers; k++) {
                            alreadySolved[k][j] = true;
                        }
                    }
                }
            }
        }
        //System.out.println("enterOrig = " + enterOrig + " vs enterNew = " + enterNew);
        //System.out.println("Costs of this order for all instances: " + orderCosts);
        System.out.println("Correct costs of this order for all instances: " + checkCorrectCost);
        // checkCorrectCost contains correct costs even if not all available solvers are picked -> write this number
        this.stringBuilder.append("costs-all-instances").append(",").append(checkCorrectCost).append(System.lineSeparator());
    }

    private void computeNumberOfNeededSolversForAllInstances() {

        List<Integer> orderedSolversAllInstances = new LinkedList<>();
        orderedSolversAllInstances.addAll(this.totalOrder);
        //System.out.println(totalOrderBySolver.toString());
        //System.out.println(orderedSolversAllInstances.toString());
        List<Integer> unsolvedInstancesAll = new LinkedList<>();

        for (int i = 0; i < this.noOfAllInstances; i++) {
            unsolvedInstancesAll.add(i);
        }

        int counterSolverAllInstances = 0;
        while (!unsolvedInstancesAll.isEmpty() && !orderedSolversAllInstances.isEmpty()) {
            // get first solver and remove the now solved instances from the list unsolvedInstances
            int chosenSolver = orderedSolversAllInstances.get(0);
            counterSolverAllInstances++;
            ((LinkedList<Integer>) orderedSolversAllInstances).removeFirstOccurrence(chosenSolver);

            String chosenSolverName = null;
            for (Map.Entry<String, Integer> entry : this.solver2id.entrySet()) {
                if (chosenSolver == entry.getValue()) {
                    chosenSolverName = entry.getKey();
                }
            }

            int solverOriginalId = -1;
            for (int j = 1; j < this.costFileData.get(0).size(); j++) {     // first entry is not a solver
                if (this.costFileData.get(0).get(j).equals(chosenSolverName)) {
                    solverOriginalId = j;
                }
            }

            for (int i = 0; i < this.noOfAllInstances; i++) {
                if (this.allSolved[solverOriginalId - 1][i] == 1) {    // -1 because index = id - 1
                    if (unsolvedInstancesAll.contains(i)) {
                        ((LinkedList<Integer>) unsolvedInstancesAll).removeFirstOccurrence(i);
                    }
                }
            }
        }

        stringBuilder.append("number-needed-solvers-all-instances").append(",").append(counterSolverAllInstances).append(System.lineSeparator());

    }

    private void writeFile() {
        try {
            File dir = new File(resultDirectoryName);
            File file = new File(dir, this.fileName);

            BufferedWriter writer = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(file), StandardCharsets.UTF_8));
            writer.append(this.stringBuilder);
            writer.flush();

        } catch (FileNotFoundException e) {
            e.printStackTrace();
        } catch (IOException e) {
            e.printStackTrace();
        }

    }

    @Override
    public void run() {
        long startTime = System.nanoTime();

        solver2id = TestMatrixGenerator.getSolver2Id();
        instance2id = TestMatrixGenerator.getInstance2Id();

        for (int i = 0; i < this.noOfPickedSolvers; i++) {
            this.solversToSort.add(i);
        }

        List<Integer> unsolvedInstances = new LinkedList<>();
        for (int i = 0; i < this.noOfPickedInstances; i++) {
            unsolvedInstances.add(i);
        }

        // count how many solvers are needed to solve the given instances
        int counterNeededSolvers = 0;

        while (!unsolvedInstances.isEmpty()) {
            this.solversSortedByNumberOfSolvedInstances = sortSolversByNumberOfSolvedInstances(unsolvedInstances);
            //System.out.println("Sorted by number of solved instances: " + this.solversSortedByNumberOfSolvedInstances.toString());
            // get first solver and remove the now solved instances from the list unsolvedInstances
            int chosenSolver = solversSortedByNumberOfSolvedInstances.get(0);
            this.totalOrder.add(chosenSolver + 1);  // +1 because number of solvers starts with 1 and not 0 like the indices do
            counterNeededSolvers++;
            ((LinkedList<Integer>) this.solversToSort).removeFirstOccurrence(chosenSolver);
            for (int i = 0; i < this.noOfPickedInstances; i++) {
                if (this.solved[chosenSolver][i] == 1) {
                    if (unsolvedInstances.contains(i)) {
                        ((LinkedList<Integer>) unsolvedInstances).removeFirstOccurrence(i);
                    }
                }
            }
        }

        long endTime = System.nanoTime();
        double elapsedTime = (double) endTime - startTime;
        double elapsedTimeInSeconds = elapsedTime * 0.000000001;

        this.stringBuilder.append("runtime-in-nanoseconds").append(",").append(elapsedTime).append(System.lineSeparator());
        this.stringBuilder.append("runtime-in-seconds").append(",").append(elapsedTimeInSeconds).append(System.lineSeparator());

        checkAllOrdered();

        if (!randomInput) {
            for (int i : this.totalOrder) {     // total order ids and solver2id ids start both with 1
                for (Map.Entry<String, Integer> entry : solver2id.entrySet()) {
                    if (entry.getValue() == i) {
                        String solverName = entry.getKey();
                        this.totalOrderByNames.add(solverName);
                    }
                }
            }
        }


        System.out.println("Greedy total order: " + this.totalOrder.toString());

        this.stringBuilder.append("order").append(",");
        for (int i = 0; i < this.totalOrder.size(); i++) {
            if (i == (this.totalOrder.size() - 1)) {
                this.stringBuilder.append(this.totalOrder.get(i)).append(System.lineSeparator());
            } else {
                this.stringBuilder.append(this.totalOrder.get(i)).append(",");
            }
        }

        this.stringBuilder.append("number-needed-solvers").append(",").append(counterNeededSolvers).append(System.lineSeparator());

        if (randomInput) {
            computeTotalOrderCostsForRandomInput();
        } else {
            //System.out.println("Greedy total order with solver names: " + this.totalOrderByNames.toString());
            this.stringBuilder.append("order-names").append(",");
            for (int i = 0; i < this.totalOrder.size(); i++) {
                if (i == (this.totalOrder.size() - 1)) {
                    this.stringBuilder.append(this.totalOrderByNames.get(i)).append(System.lineSeparator());
                } else {
                    this.stringBuilder.append(this.totalOrderByNames.get(i)).append(",");
                }
            }
            computeTotalOrderCosts();

            if (this.noOfPickedInstances < this.noOfAllInstances) {
                computeCostsOfThisOrderForAllInstances();
                computeNumberOfNeededSolversForAllInstances();
            }
        }

        System.out.println();

        writeFile();
    }

    public void startFromParser() {
        long startTime = System.nanoTime();

        //solver2id and instance2id are set by parser with setter

        for (int i = 0; i < this.noOfPickedSolvers; i++) {
            this.solversToSort.add(i);
        }

        List<Integer> unsolvedInstances = new LinkedList<>();
        for (int i = 0; i < this.noOfPickedInstances; i++) {
            unsolvedInstances.add(i);
        }

        // count how many solvers are needed to solve the given instances
        int counterNeededSolvers = 0;

        while (!unsolvedInstances.isEmpty()) {
            this.solversSortedByNumberOfSolvedInstances = sortSolversByNumberOfSolvedInstances(unsolvedInstances);
            //System.out.println("Sorted by number of solved instances: " + this.solversSortedByNumberOfSolvedInstances.toString());
            // get first solver and remove the now solved instances from the list unsolvedInstances
            int chosenSolver = solversSortedByNumberOfSolvedInstances.get(0);
            this.totalOrder.add(chosenSolver + 1);  // +1 because number of solvers starts with 1 and not 0 like the indices do
            counterNeededSolvers++;
            ((LinkedList<Integer>) this.solversToSort).removeFirstOccurrence(chosenSolver);
            for (int i = 0; i < this.noOfPickedInstances; i++) {
                if (this.solved[chosenSolver][i] == 1) {
                    if (unsolvedInstances.contains(i)) {
                        ((LinkedList<Integer>) unsolvedInstances).removeFirstOccurrence(i);
                    }
                }
            }
        }

        long endTime = System.nanoTime();
        double elapsedTime = (double) endTime - startTime;
        double elapsedTimeInSeconds = elapsedTime * 0.000000001;

        this.stringBuilder.append("runtime-in-nanoseconds").append(",").append(elapsedTime).append(System.lineSeparator());
        this.stringBuilder.append("runtime-in-seconds").append(",").append(elapsedTimeInSeconds).append(System.lineSeparator());

        checkAllOrdered();

        if (!randomInput) {
            for (int i : this.totalOrder) {     // total order ids and solver2id ids start both with 1
                for (Map.Entry<String, Integer> entry : solver2id.entrySet()) {
                    if (entry.getValue() == i) {
                        String solverName = entry.getKey();
                        this.totalOrderByNames.add(solverName);
                    }
                }
            }
        }


        System.out.println("Greedy total order: " + this.totalOrder.toString());

        this.stringBuilder.append("order").append(",");
        for (int i = 0; i < this.totalOrder.size(); i++) {
            if (i == (this.totalOrder.size() - 1)) {
                this.stringBuilder.append(this.totalOrder.get(i)).append(System.lineSeparator());
            } else {
                this.stringBuilder.append(this.totalOrder.get(i)).append(",");
            }
        }

        this.stringBuilder.append("number-needed-solvers").append(",").append(counterNeededSolvers).append(System.lineSeparator());

        if (randomInput) {
            computeTotalOrderCostsForRandomInput();
        } else {
            //System.out.println("Greedy total order with solver names: " + this.totalOrderByNames.toString());
            this.stringBuilder.append("order-names").append(",");
            for (int i = 0; i < this.totalOrder.size(); i++) {
                if (i == (this.totalOrder.size() - 1)) {
                    this.stringBuilder.append(this.totalOrderByNames.get(i)).append(System.lineSeparator());
                } else {
                    this.stringBuilder.append(this.totalOrderByNames.get(i)).append(",");
                }
            }
            computeTotalOrderCosts();

            if (this.noOfPickedInstances < this.noOfAllInstances) {
                computeCostsOfThisOrderForAllInstances();
                computeNumberOfNeededSolversForAllInstances();
            }
        }

        System.out.println();

        writeFile();
    }


    public long[][] getSolved() {
        return solved;
    }

    public void setSolved(long[][] solved) {
        this.solved = solved;
    }

    public long[][] getCosts() {
        return costs;
    }

    public void setCosts(long[][] costs) {
        this.costs = costs;
    }


    public int getNoOfPickedSolvers() {
        return noOfPickedSolvers;
    }

    public void setNoOfPickedSolvers(int noOfPickedSolvers) {
        this.noOfPickedSolvers = noOfPickedSolvers;
    }

    public int getNoOfPickedInstances() {
        return noOfPickedInstances;
    }

    public void setNoOfPickedInstances(int noOfPickedInstances) {
        this.noOfPickedInstances = noOfPickedInstances;
    }

    public boolean isRandomInput() {
        return randomInput;
    }

    public void setRandomInput(boolean randomInput) {
        this.randomInput = randomInput;
    }

    public SortedMap<String, Integer> getSolver2id() {
        return solver2id;
    }

    public void setSolver2id(SortedMap<String, Integer> solver2id) {
        this.solver2id = solver2id;
    }

    public SortedMap<String, Integer> getInstance2id() {
        return instance2id;
    }

    public void setInstance2id(SortedMap<String, Integer> instance2id) {
        this.instance2id = instance2id;
    }
}
