import java.io.*;
import java.nio.charset.StandardCharsets;
import java.util.*;

public class TestMatrixGenerator {

    private String resultDirectoryName;

    private long[][] solveMatrix;
    private long[][] costMatrix;
    private float[][] floatCostMatrix;

    private static SortedMap<String, Integer> solver2Id = new TreeMap<>();
     private static SortedMap<String, Integer> instance2Id = new TreeMap<>();

    TestMatrixGenerator(String resultDirectoryName) {
        this.resultDirectoryName = resultDirectoryName;
    }

    public void generateMatrices(int noOfAlgorithms, int noOfInstances) {
        this.solveMatrix = new long[noOfAlgorithms][noOfInstances];
        this.costMatrix = new long[noOfAlgorithms][noOfInstances];
        this.floatCostMatrix = new float[noOfAlgorithms][noOfInstances];

        Random random = new Random();

        for (int i = 0; i < noOfInstances; i++) {   // for each instance there should be at least one algorithm to solve it
            boolean solved = false;
            for (int j = 0; j < noOfAlgorithms; j++) {
                if (!solved && j == (noOfAlgorithms - 1)) {
                    this.solveMatrix[j][i] = 1;
                    this.costMatrix[j][i] = (long) random.nextInt(10) + 1;
                    this.floatCostMatrix[j][i] = this.costMatrix[j][i];
                } else {
                    int entry = random.nextInt(2);
                    this.solveMatrix[j][i] = entry;
                    if (entry == 1) {
                        solved = true;
                        this.costMatrix[j][i] = (long) random.nextInt(10) + 1; // +1 so that the minimum value is 1
                        this.floatCostMatrix[j][i] = this.costMatrix[j][i];
                    } else {    // if the algorithm does not solve the instance put a timeout of 10
                        this.costMatrix[j][i] = 10;
                        this.floatCostMatrix[j][i] = this.costMatrix[j][i];

                    }
                }

            }
        }

    }

    // optimal order of algorithms: 1 - 3 - 4 - 2 (or with index: 0 - 2 - 3 - 1)
    public void setTestInstance1() {
        this.solveMatrix = new long[4][5];
        this.costMatrix = new long[4][5];
        this.floatCostMatrix = new float[4][5];

        this.solveMatrix[0][0] = 1;
        this.solveMatrix[0][1] = 1;
        this.solveMatrix[0][2] = 1;
        this.solveMatrix[0][3] = 0;
        this.solveMatrix[0][4] = 1;

        this.solveMatrix[1][0] = 0;
        this.solveMatrix[1][1] = 0;
        this.solveMatrix[1][2] = 1;
        this.solveMatrix[1][3] = 0;
        this.solveMatrix[1][4] = 1;

        this.solveMatrix[2][0] = 0;
        this.solveMatrix[2][1] = 1;
        this.solveMatrix[2][2] = 0;
        this.solveMatrix[2][3] = 1;
        this.solveMatrix[2][4] = 0;

        this.solveMatrix[3][0] = 1;
        this.solveMatrix[3][1] = 1;
        this.solveMatrix[3][2] = 1;
        this.solveMatrix[3][3] = 0;
        this.solveMatrix[3][4] = 1;


        this.costMatrix[0][0] = 1;
        this.costMatrix[0][1] = 5;
        this.costMatrix[0][2] = 2;
        this.costMatrix[0][3] = 7;
        this.costMatrix[0][4] = 3;

        this.costMatrix[1][0] = 7;
        this.costMatrix[1][1] = 7;
        this.costMatrix[1][2] = 3;
        this.costMatrix[1][3] = 7;
        this.costMatrix[1][4] = 5;

        this.costMatrix[2][0] = 7;
        this.costMatrix[2][1] = 1;
        this.costMatrix[2][2] = 7;
        this.costMatrix[2][3] = 6;
        this.costMatrix[2][4] = 7;

        this.costMatrix[3][0] = 2;
        this.costMatrix[3][1] = 5;
        this.costMatrix[3][2] = 3;
        this.costMatrix[3][3] = 7;
        this.costMatrix[3][4] = 4;

        this.floatCostMatrix[0][0] = 1;
        this.floatCostMatrix[0][1] = 5;
        this.floatCostMatrix[0][2] = 2;
        this.floatCostMatrix[0][3] = 7;
        this.floatCostMatrix[0][4] = 3;

        this.floatCostMatrix[1][0] = 7;
        this.floatCostMatrix[1][1] = 7;
        this.floatCostMatrix[1][2] = 3;
        this.floatCostMatrix[1][3] = 7;
        this.floatCostMatrix[1][4] = 5;

        this.floatCostMatrix[2][0] = 7;
        this.floatCostMatrix[2][1] = 1;
        this.floatCostMatrix[2][2] = 7;
        this.floatCostMatrix[2][3] = 6;
        this.floatCostMatrix[2][4] = 7;

        this.floatCostMatrix[3][0] = 2;
        this.floatCostMatrix[3][1] = 5;
        this.floatCostMatrix[3][2] = 3;
        this.floatCostMatrix[3][3] = 7;
        this.floatCostMatrix[3][4] = 4;

    }

    public void setTestInstance2() {
        this.solveMatrix = new long[4][5];
        this.costMatrix = new long[4][5];
        this.floatCostMatrix = new float[4][5];

        this.solveMatrix[0][0] = 1;
        this.solveMatrix[0][1] = 0;
        this.solveMatrix[0][2] = 1;
        this.solveMatrix[0][3] = 0;
        this.solveMatrix[0][4] = 0;

        this.solveMatrix[1][0] = 0;
        this.solveMatrix[1][1] = 0;
        this.solveMatrix[1][2] = 1;
        this.solveMatrix[1][3] = 0;
        this.solveMatrix[1][4] = 1;

        this.solveMatrix[2][0] = 1;
        this.solveMatrix[2][1] = 1;
        this.solveMatrix[2][2] = 1;
        this.solveMatrix[2][3] = 0;
        this.solveMatrix[2][4] = 0;

        this.solveMatrix[3][0] = 1;
        this.solveMatrix[3][1] = 0;
        this.solveMatrix[3][2] = 1;
        this.solveMatrix[3][3] = 1;
        this.solveMatrix[3][4] = 0;


        this.costMatrix[0][0] = 2;
        this.costMatrix[0][1] = 9;
        this.costMatrix[0][2] = 1;
        this.costMatrix[0][3] = 9;
        this.costMatrix[0][4] = 9;

        this.costMatrix[1][0] = 9;
        this.costMatrix[1][1] = 9;
        this.costMatrix[1][2] = 5;
        this.costMatrix[1][3] = 9;
        this.costMatrix[1][4] = 7;

        this.costMatrix[2][0] = 6;
        this.costMatrix[2][1] = 6;
        this.costMatrix[2][2] = 4;
        this.costMatrix[2][3] = 9;
        this.costMatrix[2][4] = 9;

        this.costMatrix[3][0] = 7;
        this.costMatrix[3][1] = 9;
        this.costMatrix[3][2] = 8;
        this.costMatrix[3][3] = 8;
        this.costMatrix[3][4] = 9;
    }

    public void pickRandomlyFromGivenInput(int noOfAlgorithms, int noOfInstances, List<List<String>> costFileData, List<List<String>> solveFileData, String uniqueTime) {
        this.solveMatrix = new long[noOfAlgorithms][noOfInstances];
        this.costMatrix = new long[noOfAlgorithms][noOfInstances];
        this.floatCostMatrix = new float[noOfAlgorithms][noOfInstances];

        int noOfPossibleSolvers = solveFileData.get(0).size() - 1;  // -1 because first entry is "instance"
        int noOfPossibleInstances = solveFileData.size() - 1;       // -1 because first row is solver names / first entry is "instance"

        List<String> pickedSolver = new LinkedList<>();
        List<String> pickedInstances = new LinkedList<>();

        Random random = new Random();

        SortedMap<String, Integer> solverOriginalIndex = new TreeMap<>();
        SortedMap<String, Integer> instanceOriginalIndex = new TreeMap<>();

        if (noOfAlgorithms > noOfPossibleSolvers) {
            System.exit(-1);
        }

        if (noOfAlgorithms == noOfPossibleSolvers) {
            for (int i = 0; i < noOfPossibleSolvers; i++) {
                String solverName = solveFileData.get(0).get(i + 1);    // index 0 is not a solver
                pickedSolver.add(solverName);
                solverOriginalIndex.put(solverName, i + 1);
                solver2Id.put(solverName, pickedSolver.size());         // start ids with 1
            }

            // pick given number of instances
            int counterInstances = 0;
            while (counterInstances < noOfInstances){
                int toPick = random.nextInt(noOfPossibleInstances) + 1; // upper bound is exclusive and first entry is not a solver
                String instanceName = solveFileData.get(toPick).get(0);
                if (!pickedInstances.contains(instanceName)) {
                    // filter instances which cannot be solved by any of the chosen solvers
                    boolean isSolved = false;
                    for (int i = 0; i < pickedSolver.size(); i++) {
                        int solverIndex = solverOriginalIndex.get(pickedSolver.get(i));
                        if (Integer.parseInt(solveFileData.get(toPick).get(solverIndex)) == 1) {
                            isSolved = true;
                        }
                    }

                    if (isSolved ) {
                        pickedInstances.add(instanceName);
                        instanceOriginalIndex.put(instanceName, toPick);
                        instance2Id.put(instanceName, pickedInstances.size());
                        counterInstances++;
                    }

                }
            }

        } else {
            // pick given number of solvers
            int counterSolvers = 0;
            while (counterSolvers < noOfAlgorithms){
                int toPick = random.nextInt(noOfPossibleSolvers) + 1; // upper bound is exclusive and first entry is not a solver
                String solverName = solveFileData.get(0).get(toPick);
                if (!pickedSolver.contains(solverName)) {
                    pickedSolver.add(solverName);
                    solverOriginalIndex.put(solverName, toPick);
                    solver2Id.put(solverName, pickedSolver.size());
                    counterSolvers++;
                }
            }

            // pick given number of instances
            int counterInstances = 0;
            while (counterInstances < noOfInstances){
                int toPick = random.nextInt(noOfPossibleInstances) + 1; // upper bound is exclusive and first entry is not a solver
                String instanceName = solveFileData.get(toPick).get(0);
                if (!pickedInstances.contains(instanceName)) {
                    // filter instances which cannot be solved by any of the chosen solvers
                    boolean isSolved = false;
                    for (int i = 0; i < pickedSolver.size(); i++) {
                        int solverIndex = solverOriginalIndex.get(pickedSolver.get(i));
                        if (Integer.parseInt(solveFileData.get(toPick).get(solverIndex)) == 1) {
                            isSolved = true;
                        }
                    }

                    if (isSolved ) {
                        pickedInstances.add(instanceName);
                        instanceOriginalIndex.put(instanceName, toPick);
                        instance2Id.put(instanceName, pickedInstances.size());
                        counterInstances++;
                    }

                }
            }
        }




        for (int i = 0; i < noOfAlgorithms; i++) {
            int algIndex = solverOriginalIndex.get(pickedSolver.get(i));

            for (int j = 0; j < noOfInstances; j++) {
                int instanceIndex = instanceOriginalIndex.get(pickedInstances.get(j));

                this.solveMatrix[i][j] = Long.parseLong(solveFileData.get(instanceIndex).get(algIndex));
                this.costMatrix[i][j] = (long)(Double.parseDouble(costFileData.get(instanceIndex).get(algIndex)) * 1000);    // convert cut time to milliseconds!! -> weight must be an int/long greater than 0!
                this.floatCostMatrix[i][j] = Float.parseFloat(costFileData.get(instanceIndex).get(algIndex));
            }
        }

        String solverMappingFileName = "solver2id-" + uniqueTime + ".csv";
        String instanceMappingFileName = "instance2id-" + uniqueTime + ".csv";

        try {
            File dir = new File(this.resultDirectoryName);
            writeSolverMapping(new File(dir, solverMappingFileName), solver2Id);
            writeInstanceMapping(new File(dir, instanceMappingFileName), instance2Id);
        } catch (IOException e) {
            e.printStackTrace();
        }


    }

    private static void writeSolverMapping(File mappingFile, SortedMap<String, Integer> solver2Id) throws IOException {
        StringBuilder sb = new StringBuilder();
        for (Map.Entry<String, Integer> entry : solver2Id.entrySet()) {
            sb.append(entry.getKey()).append(",").append(entry.getValue()).append(System.lineSeparator());
        }
        try (BufferedWriter writer = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(mappingFile), StandardCharsets.UTF_8))) {
            writer.append(sb);
            writer.flush();
        }
    }

    private static void writeInstanceMapping(File mappingFile, SortedMap<String, Integer> instance2Id) throws IOException {
        StringBuilder sb = new StringBuilder();
        for (Map.Entry<String, Integer> entry : instance2Id.entrySet()) {
            sb.append(entry.getKey()).append(",").append(entry.getValue()).append(System.lineSeparator());
        }
        try (BufferedWriter writer = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(mappingFile), StandardCharsets.UTF_8))) {
            writer.append(sb);
            writer.flush();
        }
    }

    public long[][] getSolveMatrix() {
        return solveMatrix;
    }

    public void setSolveMatrix(long[][] solveMatrix) {
        this.solveMatrix = solveMatrix;
    }

    public long[][] getCostMatrix() {
        return costMatrix;
    }

    public void setCostMatrix(long[][] costMatrix) {
        this.costMatrix = costMatrix;
    }

    public float[][] getFloatCostMatrix() {
        return floatCostMatrix;
    }

    public void setFloatCostMatrix(float[][] floatCostMatrix) {
        this.floatCostMatrix = floatCostMatrix;
    }

    public static SortedMap<String, Integer> getSolver2Id() {
        return solver2Id;
    }

    public static void setSolver2Id(SortedMap<String, Integer> solver2Id) {
        solver2Id = solver2Id;
    }

    public static SortedMap<String, Integer> getInstance2Id() {
        return instance2Id;
    }

    public static void setInstance2Id(SortedMap<String, Integer> instance2Id) {
        instance2Id = instance2Id;
    }
}
